class PhysWorld
{
  Vector gravity;
  int num_objects;
  PhysObject[] objects;
  int num_links;
  PhysLink[] links;
  int num_blacklist;
  Pointer[][] colBlackList;

  PhysWorld()
  {
    num_objects = 0;
    num_links = 0;
    num_blacklist = 0;
    gravity = new Vector();
    objects = new PhysObject[0];
    colBlackList = new Pointer[0][2];
  }

  //==================================
  // UPDATE PHYSICS WORLD
  //==================================

  void update()
  {       
    float num_step = 4;
    
    if (millis() > 2000)
    {
      for (int k=0; k<num_step; k++)
      {
        float delta_t = 1/FRS.value()/num_step;
        
        for (int i=0; i<num_objects; i++)
        {
          objects[i].reset();     
        }
        
        for (int i=0; i<num_links; i++)
        {
          links[i].update();
        }
        
        for (int i=0; i<num_objects; i++)
        { 
          objects[i].update(gravity,delta_t);    
          for (int j=i+1; j<num_objects; j++)
          {            
            if (!blackListed(new Pointer(i),new Pointer(j)))
            {
              collide(objects[i], objects[j]);
            }
          }      
        }
      }
    }
  }

  //==================================
  // COLLISION DETECTION AND RESPONSE
  //==================================

  void collide(PhysObject A, PhysObject B)
  {
    float delta_x = A.center.x - B.center.x;
    float delta_y = A.center.y - B.center.y;
    float delta_d = sqrt(pow(delta_x,2)+pow(delta_y,2));
    float sum_r = A.poly.radius + B.poly.radius;
            
    if ((A.material.invDensity != 0 || B.material.invDensity != 0) && delta_d <= sum_r)
    {
      Vector MTD = calculateMTD(A, B);
       
      if ( MTD.len != 0 )
      {
        Point[] contactPoints = findContactPoints(A,B,MTD);
        
        Vector vA = A.veloc;
        Vector vB = B.veloc;
        Vector vAB = vA.addVector(vB.multiply(-1));
        Vector rA = new Vector(contactPoints[0]);
        Vector rB = new Vector(contactPoints[1]);
        
        float e = A.material.bounce*B.material.bounce;
        float invMassA = A.material.invDensity/A.poly.area;
        float invMassB = B.material.invDensity/B.poly.area;
               
        Vector vtA = A.veloc.addVector(rA.multiply(A.rotationSpeed));
        Vector vtB = B.veloc.addVector(rB.multiply(B.rotationSpeed));
        Vector vtAB = vA.addVector(vB.multiply(-1));       

        float j = (-(1+e)*vAB.dot(MTD))/(MTD.dot(MTD)*(invMassA+invMassB)+pow(rA.normal().dot(MTD),2)*A.invMomentOfInertia+pow(rB.normal().dot(MTD),2)*B.invMomentOfInertia);
        
        if (invMassA != 0) A.impulse(MTD,+0.5,+j,rA);
        if (invMassB != 0) B.impulse(MTD,-0.5,-j,rB);
        
        float frictionK = A.material.frictionK*B.material.frictionK;
        Vector frictionV = new Vector();
        frictionV.dir = MTD.normal().dir;
        frictionV.len = vtAB.dot(MTD.normal())*frictionK*j;
        
        // Hack to prevent very large friction forces - Worst hack of them all
        if (abs(frictionV.len)>1E4) frictionV.len *= 1E4/abs(frictionV.len);
        // End hack
        
        if (vtAB.dot(MTD) <= 0)
        {
          A.addForce(rA.getCoords(), frictionV.multiply(-1));
          B.addForce(rB.getCoords(), frictionV.multiply(+1));         
        }
      }
    }
  }
  
  Point[] findContactPoints(PhysObject A, PhysObject B, Vector MTD)
  {
    Point[] R = new Point[2];
    R[0] = new Point();
    R[1] = new Point();
    
    Point[] supportA = findSupportPoints(A.poly, MTD);
    Point[] supportB = findSupportPoints(B.poly, MTD.multiply(-1));

    int sNumA = numSupportPoints(supportA);
    int sNumB = numSupportPoints(supportB);

    if (sNumA == 1 && sNumB == 1)
    {
      R[0] = supportA[0];
      R[1] = supportB[0];
    }
    else if (sNumA == 1 && sNumB == 2)
    {
      R[0] = supportA[0];
      Vector edge = new Vector(supportB[1].x-supportB[0].x, supportB[1].y-supportB[0].y);
      R[1] = projectPointOnEdge(A.center.addVector(new Vector(supportA[0])), edge, B.center.addVector(new Vector(supportB[0]))).addVector(new Vector(B.center.multiply(-1)));
    }
    else if (sNumA == 2 && sNumB == 1)
    {
      R[1] = supportB[0];
      Vector edge = new Vector(supportA[1].x-supportA[0].x, supportA[1].y-supportA[0].y);
      R[0] = projectPointOnEdge(B.center.addVector(new Vector(supportB[0])), edge, A.center.addVector(new Vector(supportA[0]))).addVector(new Vector(A.center.multiply(-1)));
    }
    else if (sNumA == 2 && sNumB == 2)
    {
      // Hack to prevent objects from shaking on top of each other
      A.rotationSpeed = 0;
      B.rotationSpeed = 0;
      // End hack
      
      Point[] sortedPoints = new Point[4];
      String[] sortedPointsType = new String[4];
      
      sortedPoints[0] = supportA[0];
      sortedPoints[1] = supportA[1];
      sortedPoints[2] = supportB[0];
      sortedPoints[3] = supportB[1];
      
      sortedPointsType[0] = "A";
      sortedPointsType[1] = "A";
      sortedPointsType[2] = "B";
      sortedPointsType[3] = "B";
      
      Vector normalMTD = MTD.normal();
      
      for (int i=3; i>=0; i--)
      {
        for (int j=0; j<=i-1; j++)
        {
          Point centerJ = new Point();
          Point centerJPlus1 = new Point();
          
          if ( sortedPointsType[j] == "A" ) 
          {
            centerJ = A.center;
          } else {
            centerJ = B.center; 
          }
          if ( sortedPointsType[j+1] == "A" ) {
            centerJPlus1 = A.center;
          } else {
            centerJPlus1 = B.center; 
          }
          
          float dotJ = normalMTD.dot(new Vector(sortedPoints[j])) + normalMTD.dot(new Vector(centerJ));
          float dotJPlus1 = normalMTD.dot(new Vector(sortedPoints[j+1])) + normalMTD.dot(new Vector(centerJPlus1));
          
          if ( dotJ > dotJPlus1 )
          {
            Point T1 = sortedPoints[j];
            sortedPoints[j] = sortedPoints[j+1];
            sortedPoints[j+1] = T1;
            
            String T2 = sortedPointsType[j];
            sortedPointsType[j] = sortedPointsType[j+1];
            sortedPointsType[j+1] = T2;
          }
        }
      }
      
      if ( sortedPointsType[1] == "B" && sortedPointsType[2] == "A" )
      {
        R[0] = projectPointOnEdge(B.center.addVector(new Vector(sortedPoints[1])), normalMTD, A.center.addVector(new Vector(sortedPoints[2]))).addVector(new Vector(A.center.multiply(-1)));
        R[1] = projectPointOnEdge(A.center.addVector(new Vector(sortedPoints[2])), normalMTD, B.center.addVector(new Vector(sortedPoints[1]))).addVector(new Vector(B.center.multiply(-1)));
      }
      else if ( sortedPointsType[1] == "A" && sortedPointsType[2] == "B" )
      {
        R[0] = projectPointOnEdge(B.center.addVector(new Vector(sortedPoints[2])), normalMTD, A.center.addVector(new Vector(sortedPoints[1]))).addVector(new Vector(A.center.multiply(-1)));
        R[1] = projectPointOnEdge(A.center.addVector(new Vector(sortedPoints[1])), normalMTD, B.center.addVector(new Vector(sortedPoints[2]))).addVector(new Vector(B.center.multiply(-1)));
      }
      else if ( sortedPointsType[1] == "A" && sortedPointsType[2] == "A" )
      {
        Point middlePoint = sortedPoints[1].addVector(new Vector(sortedPoints[2])).multiply(0.5);
        R[0] = middlePoint;
        R[1] = projectPointOnEdge(A.center.addVector(new Vector(middlePoint)), normalMTD, B.center.addVector(new Vector(supportB[0]))).addVector(new Vector(B.center.multiply(-1)));
      }
      else if ( sortedPointsType[1] == "B" && sortedPointsType[2] == "B" )
      {
        Point middlePoint = sortedPoints[1].addVector(new Vector(sortedPoints[2])).multiply(0.5);
        R[0] = projectPointOnEdge(B.center.addVector(new Vector(middlePoint)), normalMTD, A.center.addVector(new Vector(supportA[0]))).addVector(new Vector(A.center.multiply(-1)));
        R[1] = middlePoint;        
      }
    }
    
    return R;
  }
   
  Point[] findSupportPoints(Polygon P, Vector V)
  {
    Point supportPoints[] = new Point[2];
    supportPoints[0] = P.vertices[0];
    supportPoints[1] = new Point(0,0);
    float minimumDot = new Vector(P.vertices[0]).dot(V);
    float[] dot = new float[P.vertices.length];
    dot[0] = minimumDot;
    
    for (int i=1; i<P.vertices.length; i++)
    {
      dot[i] = new Vector(P.vertices[i]).dot(V);
      if (dot[i] < minimumDot)
      {
        minimumDot = dot[i];
      }
    }

    final float threshold = 1E-6;
    int index = 0;
    
    for (int i=0; i<P.vertices.length; i++)
    {
      if (dot[i]-minimumDot < threshold)
      {
        supportPoints[index++] = P.vertices[i];
        if (index==2) break;
      }
    }

    return supportPoints;
  }
  
  int numSupportPoints(Point[] support)
  {
    int R = 1;
    if ( support[1].x != 0 || support[1].y != 0 ) R++;
    return R;
  }
  
  Point projectPointOnEdge(Point P, Vector V, Point offsetV)
  {
    Point R = new Point();
    
    Vector vectorP = new Vector(P);
    Vector vectorR = new Vector();
    vectorP = vectorP.addVector(new Vector(offsetV).multiply(-1));
    vectorR.dir = V.dir;
    vectorR.len = V.dot(vectorP)/V.len;
    R = vectorR.addVector(new Vector(offsetV)).getCoords();
    
    return R;
  }
  
  Vector calculateMTD(PhysObject A, PhysObject B)
  {
    int index = 0;
    Vector[] pushVectors = new Vector[A.poly.numEdges()+B.poly.numEdges()];

    for (int i = 0; i<A.poly.numEdges(); i++)
    {
      float intervalIntersection = calculateIntervalIntersection(A.poly.getEdge(i).normal(), A, B);
      if (intervalIntersection == 0)
      {
        return new Vector();
      }
      pushVectors[index] = new Vector();
      pushVectors[index].len = intervalIntersection;
      pushVectors[index].dir = A.poly.getEdge(i).normal().dir;
      index++;
    } 
    for (int i = 0; i<B.poly.numEdges(); i++)
    {       
      float intervalIntersection = calculateIntervalIntersection(B.poly.getEdge(i).normal(), A, B);
      if (intervalIntersection == 0)
      {
        return new Vector();
      }
      pushVectors[index] = new Vector();
      pushVectors[index].len = intervalIntersection;
      pushVectors[index].dir = B.poly.getEdge(i).normal().dir;
      index++;
    }

    Vector currentBest = new Vector();

    for (int i=0; i<pushVectors.length; i++)
    {
      if (( pushVectors[i].len < currentBest.len && pushVectors[i].len != 0 ) || currentBest.len == 0 )
      {
        currentBest = pushVectors[i];
      }
    }

    Vector D = new Vector(A.center.addVector(new Vector(B.center.multiply(-1)))); 
    if (D.dot(currentBest) < 0)
    { 
      currentBest = currentBest.multiply(-1); 
    }

    return currentBest; 
  }

  float calculateIntervalIntersection(Vector N, PhysObject A, PhysObject B)
  {
    Interval intervalA = calculateInterval(N,A).offset(new Vector(A.center.x,A.center.y).dot(N)/N.len);
    Interval intervalB = calculateInterval(N,B).offset(new Vector(B.center.x,B.center.y).dot(N)/N.len);
    Interval intersection = intervalA.intersect(intervalB);            
    return intersection.length();
  }

  Interval calculateInterval(Vector axis, PhysObject O)
  { 
    Interval R = new Interval(0,0);
    for (int i=0; i<O.poly.numEdges(); i++)
    {
      R.insert((O.poly.getEdge(i).dot(axis) + new Vector(O.poly.getEdgeStart(i)).dot(axis))/axis.len);
    }
    return R;
  } 

  //==================================
  // INTERFACE
  //==================================

  void setGravity(float len, float dir)
  {
    gravity.len = len;
    gravity.dir = dir;
  }

  Pointer addObject(float x, float y, Material m, Polygon p)
  {
    num_objects++;
    PhysObject[] temp = new PhysObject[num_objects];

    for (int i=0; i<num_objects-1; i++)
    {
      temp[i] = objects[i];
    }

    temp[num_objects-1] = new PhysObject();
    temp[num_objects-1].center = new Point(x,y);
    temp[num_objects-1].material = m;
    temp[num_objects-1].poly = p;
    temp[num_objects-1].calculateMoment();
    objects = temp;

    return new Pointer(num_objects-1);
  }
  
  void addLink(Pointer A, Pointer AWeld, Pointer B, Pointer BWeld, float len, float k)
  {
    num_links++;
    PhysLink[] temp = new PhysLink[num_links];

    for (int i=0; i<num_links-1; i++)
    {
      temp[i] = links[i];
    }

    temp[num_links-1] = new PhysLink(A, B, AWeld, BWeld, len, k);
    links = temp;
  }
  
  void popLink()
  {
    num_links--;
    PhysLink[] temp = new PhysLink[num_links];

    for (int i=0; i<num_links; i++)
    {
      temp[i] = links[i];
    }

    links = temp;
  }

  Pointer addWeldPoint(Pointer o, Point p)
  {
    for (int i=0; i<objects[o.index].weldPoints.length; i++)
    {
      if ( objects[o.index].weldPoints[i].x == -1.0 && objects[o.index].weldPoints[i].y == -1.0 )
      {
        objects[o.index].weldPoints[i] = p;
        return new Pointer(i);
      }
    }
    
    return new Pointer(objects[o.index].weldPoints.length);
  }
  
  void popWeld(Pointer o)
  {
    for (int i=objects[o.index].weldPoints.length-1; i>=0; i--)
    {
      if ( objects[o.index].weldPoints[i].x != -1.0 && objects[o.index].weldPoints[i].y != -1.0 )
      {
        objects[o.index].weldPoints[i] = new Point(-1,-1);
        break;
      }
    }
  }

  void setRotationSpeed(Pointer o, float w)
  {
    objects[o.index].rotationSpeed = w;
  }

  void setRotation(Pointer o, float r)
  {
    objects[o.index].poly.rotatePoly(r);
  }
  
  void setPosition(Pointer o, Point p)
  {
    objects[o.index].center = p;
    objects[o.index].veloc.len = 0;
    objects[o.index].rotationSpeed = 0;
  }
  
  void addColBlackList(Pointer p1, Pointer p2)
  {
    num_blacklist++;
    Pointer[][] temp = new Pointer[num_blacklist][2];

    for (int i=0; i<num_blacklist-1; i++)
    {
      temp[i] = colBlackList[i];
    }

    temp[num_blacklist-1][0] = p1;
    temp[num_blacklist-1][1] = p2;
    colBlackList = temp;
  }
  
  boolean blackListed(Pointer p1, Pointer p2)
  {
    for (int i=0; i<num_blacklist; i++)
    {
      if ((colBlackList[i][0].index == p1.index && colBlackList[i][1].index == p2.index) || (colBlackList[i][0].index == p2.index && colBlackList[i][1].index == p1.index) || 
          (colBlackList[i][0].index == p1.index && colBlackList[i][1].index == -1) || (colBlackList[i][0].index == p2.index && colBlackList[i][1].index == -1))
      {
        return true;
      }
    }
    
    return false;
  }
  
  Pointer getObjAt(Point p)
  {
    for (int i=0; i<num_objects; i++)
    {
      if (objects[i].inObj(p))
      {
        return new Pointer(i);
      }
    }
    
    return new Pointer(-1);
  }
}
